Package: h5mread
Title: A fast HDF5 reader
Description: The main function in the h5mread package is h5mread(), which
	allows reading arbitrary data from an HDF5 dataset into R, similarly
	to what the h5read() function from the rhdf5 package does.
	In the case of h5mread(), the implementation has been optimized to
	make it as fast and memory-efficient as possible.
biocViews: Infrastructure, DataRepresentation, DataImport
URL: https://bioconductor.org/packages/h5mread
BugReports: https://github.com/Bioconductor/h5mread/issues
Version: 1.4.1
License: Artistic-2.0
Encoding: UTF-8
Authors@R: person("Hervé", "Pagès", role=c("aut", "cre"),
	          email="hpages.on.github@gmail.com",
		  comment=c(ORCID="0009-0002-8272-4522"))
Depends: R (>= 4.5), methods, rhdf5, BiocGenerics, SparseArray
Imports: stats, tools, rhdf5filters, S4Vectors (>= 0.50.2), IRanges,
        S4Arrays
LinkingTo: Rhdf5lib, S4Vectors
SystemRequirements: GNU make
Suggests: BiocParallel, ExperimentHub, TENxBrainData, HDF5Array,
        testthat, knitr, rmarkdown, BiocStyle
VignetteBuilder: knitr
Collate: utils.R h5dim.R H5File-class.R h5ls.R H5DSetDescriptor-class.R
        uaselection.R h5mread.R h5summarize.R h5mread_from_reshaped.R
        h5dimscales.R h5writeDimnames.R zzz.R
git_url: https://git.bioconductor.org/packages/h5mread
git_branch: RELEASE_3_23
git_last_commit: 13dfc9b
git_last_commit_date: 2026-08-20
Repository: Bioconductor 3.23
Date/Publication: 2026-08-23
NeedsCompilation: yes
Packaged: 2026-08-23 22:30:06 UTC; biocbuild
Author: Hervé Pagès [aut, cre] (ORCID: <https://orcid.org/0009-0002-8272-4522>)
Maintainer: Hervé Pagès <hpages.on.github@gmail.com>
