# h5mread [<img src="https://bioconductor.org/images/logo/jpg/bioconductor_logo_rgb.jpg" width="200" align="right"/>](https://bioconductor.org/)

[![in_Bioc](https://bioconductor.org/shields/years-in-bioc/h5mread.svg)](https://bioconductor.org/packages/h5mread)
[![rank](https://bioconductor.org/shields/downloads/release/h5mread.svg)](https://bioconductor.org/packages/stats/bioc/h5mread)

**h5mread** is an R/Bioconductor package that allows fast and memory-efficient loading of HDF5 data into R.

See <https://bioconductor.org/packages/h5mread> for more information,
including for how to install the release version of the package (please refrain
from installing directly from GitHub).

|BioC version|Status|
|------------|------|
| release | [![build](https://bioconductor.org/shields/build/release/bioc/h5mread.svg)](https://bioconductor.org/checkResults/release/bioc-LATEST/h5mread/) [![platforms](https://bioconductor.org/shields/availability/release/h5mread.svg)](https://bioconductor.org/packages/release/h5mread#archives) [![dependencies](https://bioconductor.org/shields/dependencies/release/h5mread.svg)](https://bioconductor.org/packages/release/h5mread#since) |
| devel   | [![build](https://bioconductor.org/shields/build/devel/bioc/h5mread.svg)](https://bioconductor.org/checkResults/devel/bioc-LATEST/h5mread/) [![platforms](https://bioconductor.org/shields/availability/devel/h5mread.svg)](https://bioconductor.org/packages/devel/h5mread#archives) [![dependencies](https://bioconductor.org/shields/dependencies/devel/h5mread.svg)](https://bioconductor.org/packages/devel/h5mread#since) |

