This is the extdata/ folder of the h5mread package.

It contains the following files:

- README.txt: This file.

- test.h5: An HDF5 file containing a small collection of toy datasets used
  in the man page examples and unit tests of the package. See make_test.h5.R
  in the scripts/ folder of the h5mread package for how the file was produced.

